@@ -1,16 +1,18 @@
+"""a module to parse .ini files"""
+
 from configparser import ConfigParser, ExtendedInterpolation
 from datetime import datetime
 from functools import partial
 import sys
 import importlib.util
 from pathlib import Path
-import shutil
-from .rtfits import Fits
+from .rtfits import Fits, MultiFits
+from . import log
 
 
 class RT1_configparser(object):
-    '''
-    A configparser that can be used to fully specify a RT1-processing routine
+    """
+    A configparser that can be used to fully specify a RT1-processing routine.
 
     Extended interpolation is used -> variables can be addressed within the
     config-file by using:
@@ -24,7 +26,7 @@ class RT1_configparser(object):
         >>> var4 = ${class1:var1} bsdf
 
     Methods
-    --------
+    -------
     get_config():
         get a dict of the following structure:
 
@@ -126,17 +128,16 @@ class RT1_configparser(object):
               for example:
 
               >>> list__datetime__dts = [1.1.2018, 1.10.2018] fmt= %d%m%Y
-
-    '''
-
+    """
 
     def __init__(self, configpath, interpolation=ExtendedInterpolation()):
 
         self.configpath = Path(configpath)
         self.interpolation = interpolation
         # setup config (allow empty values -> will result in None)
-        self.config = ConfigParser(allow_no_value=True,
-                                   interpolation=self.interpolation)
+        self.config = ConfigParser(
+            allow_no_value=True, interpolation=self.interpolation
+        )
         # avoid converting uppercase characters
         # (see https://stackoverflow.com/a/19359720/9703451)
         self.config.optionxform = str
@@ -145,51 +146,63 @@ class RT1_configparser(object):
         self.cfg = self.config.read(self.configpath)
 
         # keys that will be converted to int, float or bool
-        self.lsq_parse_props = dict(section = 'least_squares_kwargs',
-                                    int_keys = ['verbose', 'max_nfev'],
-                                    float_keys = ['ftol', 'gtol', 'xtol'])
+        self.lsq_parse_props = dict(
+            section="least_squares_kwargs",
+            int_keys=["verbose", "max_nfev"],
+            float_keys=["ftol", "gtol", "xtol"],
+        )
 
-        self.fitargs_parse_props = dict(section = 'fits_kwargs',
-                                        bool_keys = ['sig0', 'dB', 'int_Q'],
-                                        int_keys= ['verbose'],
-                                        list_keys = [])
+        self.fitargs_parse_props = dict(
+            section="fits_kwargs",
+            bool_keys=["sig0", "dB", "int_Q"],
+            int_keys=["verbose"],
+            list_keys=[],
+        )
 
         self._check_integrity()
 
     def _check_integrity(self):
 
         # check if required sections are defined
-        required_sections = ['PROCESS_SPECS', 'CONFIGFILES', 'fits_kwargs',
-                             'least_squares_kwargs', 'defdict', 'RT1_V',
-                             'RT1_SRF']
+        required_sections = [
+            "PROCESS_SPECS",
+            "CONFIGFILES",
+            "fits_kwargs",
+            "least_squares_kwargs",
+            "defdict",
+            "RT1_V",
+            "RT1_SRF",
+        ]
         for key in required_sections:
-            assert key in self.config, (
-                f'a section "{key}" MUST be present in the .ini file!')
+            assert (
+                key in self.config
+            ), f'a section "{key}" MUST be present in the .ini file!'
 
         # check if save_path etc. has been defined
-        required_process_spec_keys = ['path__save_path', 'dumpfolder']
+        required_process_spec_keys = ["path__save_path", "dumpfolder"]
         for key in required_process_spec_keys:
-            assert key in self.config['PROCESS_SPECS'], (
-                f'a key "{key}" MUST be provided in the section PROCESS_SPECS')
-
+            assert (
+                key in self.config["PROCESS_SPECS"]
+            ), f'a key "{key}" MUST be provided in the section PROCESS_SPECS'
 
         # check if processing-modue is specified
-        if 'processing_cfg_module' in self.config['CONFIGFILES']:
-            proc_module_name = self.config['CONFIGFILES'
-                                           ]['processing_cfg_module']
+        if "processing_cfg_module" in self.config["CONFIGFILES"]:
+            proc_module_name = self.config["CONFIGFILES"]["processing_cfg_module"]
         else:
-            proc_module_name = 'processing_cfg'
+            proc_module_name = "processing_cfg"
 
-        assert ('module__' + proc_module_name) in self.config['CONFIGFILES'], (
-            '"module__' + proc_module_name + '" must be defined in the '
-            + 'CONFIGFILES section... \n(specify "processing_cfg_module" if ' +
-            'you want to use a different name for the processing-module key!)')
+        assert ("module__" + proc_module_name) in self.config["CONFIGFILES"], (
+            '"module__'
+            + proc_module_name
+            + '" must be defined in the '
+            + 'CONFIGFILES section... \n(specify "processing_cfg_module" if '
+            + "you want to use a different name for the processing-module key!)"
+        )
 
-
-
-    def _parse_dict(self, section, int_keys=[], float_keys=[], bool_keys=[],
-                    list_keys=[]):
-        '''
+    def _parse_dict(
+        self, section, int_keys=[], float_keys=[], bool_keys=[], list_keys=[]
+    ):
+        """
         a function to convert the parsed string values to int, float or bool
         (any additional values will be left unchanged)
 
@@ -208,8 +221,7 @@ class RT1_configparser(object):
         -------
         parsed_dict : dict
             a dict with the converted values.
-
-        '''
+        """
 
         inp = self.config[section]
 
@@ -222,21 +234,21 @@ class RT1_configparser(object):
             elif key in bool_keys:
                 val = inp.getboolean(key)
             elif key in list_keys:
-                #assert inp[key].startswith('['), f'{key}  must start with "[" '
-                #assert inp[key].endswith(']'), f'{key} must end with "]" '
-                #val = inp[key][1:-1].replace(' ', '').split(',')
-                if inp[key] is None:
+                if inp[key] == "None":
                     val = []
                 else:
-                    val = inp[key].replace(' ', '').split(',')
+                    val = inp[key].replace(" ", "").split(",")
                     val = [i for i in val if len(i) > 0]
             else:
-                val = inp[key]
+                # parse None as "real" None
+                if inp[key] == "None":
+                    val = None
+                else:
+                    val = inp[key]
 
             parsed_dict[key] = val
         return parsed_dict
 
-
     def _parse_V_SRF(self, section):
 
         inp = self.config[section]
@@ -244,20 +256,19 @@ class RT1_configparser(object):
         parsed_dict = dict()
         for key in inp:
             val = None
-            if key == 'ncoefs':
+            if key == "ncoefs":
                 val = inp.getint(key)
             else:
                 # try to convert floats, if it fails return the string
                 try:
                     val = inp.getfloat(key)
-                except:
+                except Exception:
                     val = inp[key]
 
             parsed_dict[key] = val
 
         return parsed_dict
 
-
     def _parse_defdict(self, section):
         inp = self.config[section]
 
@@ -265,82 +276,191 @@ class RT1_configparser(object):
         for key in inp:
             val = inp[key]
 
-            if val.startswith('[') and val.endswith(']'):
-                val = val[1:-1].replace(' ', '').split(',')
+            if val.startswith("[") and val.endswith("]"):
+                val = val[1:-1].replace(" ", "").split(",")
             else:
-                assert False, f'the passed defdict for {key} is not a list'
+                assert False, f"the passed defdict for {key} is not a list"
 
             # convert values
             parsed_val = []
-            if val[0] == 'True':
+            if val[0] == "True":
                 parsed_val += [True]
-            elif val[0] == 'False':
+            elif val[0] == "False":
                 parsed_val += [False]
             else:
-                assert False, (f'the passed first defdict-argument of {key} ' +
-                               f'({val[0]}) must be either "True" or "False"')
+                assert False, (
+                    f"the passed first defdict-argument of {key} "
+                    + f'({val[0]}) must be either "True" or "False"'
+                )
 
-            if val[1] == 'auxiliary':
+            if val[1] == "auxiliary":
                 parsed_val += [val[1]]
             else:
                 parsed_val += [float(val[1])]
 
             if len(val) > 2:
-                if val[2] == 'None':
+                if val[2] == "None":
                     parsed_val += [None]
                 else:
                     parsed_val += [val[2]]
 
-                parsed_val += [([float(val[3])],
-                                [float(val[4])])]
+                parsed_val += [([float(val[3])], [float(val[4])])]
 
                 try:
                     interp_Q = val[5]
-                    if interp_Q == 'True':
+                    if interp_Q == "True":
                         parsed_val += [True]
                     else:
                         parsed_val += [False]
-                except:
+                except Exception:
                     parsed_val += [False]
 
-
             parsed_dict[key] = parsed_val
 
         return parsed_dict
 
-
     def _to_dt(self, s, fmt=None):
         if fmt is None:
-            fmt = '%Y-%m-%d %H:%M:%S.%f'
+            fmt = "%Y-%m-%d %H:%M:%S.%f"
         return datetime.strptime(s, fmt)
 
+    def get_config(self, config=None):
+        """
+        get the RT1 model-configuration dict
+
+        Parameters
+        ----------
+        config : str, optional
+            The name of the config to be loaded if multiple configurations
+            are specified in the .ini file. The default is None
+
+        Returns
+        -------
+        dict
+            a dict with the model-configuration.
+        """
 
-    def get_config(self):
         lsq_kwargs = self._parse_dict(**self.lsq_parse_props)
-
         fits_kwargs = self._parse_dict(**self.fitargs_parse_props)
-        defdict = self._parse_defdict('defdict')
-        set_V_SRF = dict(V_props = self._parse_V_SRF('RT1_V'),
-                         SRF_props = self._parse_V_SRF('RT1_SRF'))
+        defdict = self._parse_defdict("defdict")
+        set_V_SRF = dict(
+            V_props=self._parse_V_SRF("RT1_V"),
+            SRF_props=self._parse_V_SRF("RT1_SRF"),
+        )
 
-        return dict(lsq_kwargs=lsq_kwargs,
-                    defdict=defdict,
-                    set_V_SRF=set_V_SRF,
-                    fits_kwargs=fits_kwargs)
+        # update configurations in case a specific config is selected
+        if config is not None:
+            # update least_squares_kwargs
+            if f"#config {config} {self.lsq_parse_props['section']}" in self.config:
+                new_lsq_parse_props = {
+                    **self.lsq_parse_props,
+                    "section": f"#config {config} {self.lsq_parse_props['section']}",
+                }
+                lsq_kwargs.update(self._parse_dict(**new_lsq_parse_props))
 
+            # update fits_kwargs
+            if f"#config {config} {self.fitargs_parse_props['section']}" in self.config:
+                new_fitargs_parse_props = {
+                    **self.fitargs_parse_props,
+                    "section": f"#config {config} {self.fitargs_parse_props['section']}",
+                }
+                fits_kwargs.update(self._parse_dict(**new_fitargs_parse_props))
+
+            # update defdict
+            if f"#config {config} defdict" in self.config:
+                defdict.update(self._parse_defdict(f"#config {config} defdict"))
+
+            # update RT1_V
+            if f"#config {config} RT1_V" in self.config:
+                V_props = self._parse_V_SRF("RT1_V")
+                V_props.update(self._parse_V_SRF(f"#config {config} RT1_V"))
+                set_V_SRF["V_props"] = V_props
+            # update RT1_SRF
+            if f"#config {config} RT1_SRF" in self.config:
+                SRF_props = self._parse_V_SRF("RT1_SRF")
+                SRF_props.update(self._parse_V_SRF(f"#config {config} RT1_SRF"))
+                set_V_SRF["SRF_props"] = SRF_props
+
+        return dict(
+            lsq_kwargs=lsq_kwargs,
+            defdict=defdict,
+            set_V_SRF=set_V_SRF,
+            fits_kwargs=fits_kwargs,
+        )
+
+    @property
+    def config_names(self):
+        config_sections = []
+        for i in self.config:
+            if i.startswith("#config"):
+                i_split = i.split(" ")
+
+                if i_split[0] == "#config":
+                    if len(i_split) == 2:
+                        config_sections.append(i_split[1])
+                    elif len(i_split) == 3:
+                        name, section = i_split[1:]
+                        assert (
+                            section in self.config
+                        ), f"the provided config section {i} is invalid"
+
+                        config_sections.append(name)
+                    else:
+                        raise AssertionError(
+                            f"there is something wrong in section {i} "
+                            + "of the .ini file"
+                        )
+
+        return set(config_sections)
 
     def get_fitobject(self):
-        cfg = self.get_config()
+        configs = self.config_names
+        if len(configs) > 0:
+            rt1_fits = MultiFits()
+            for config in configs:
+                cfg = self.get_config(config)
 
-        rt1_fits = Fits(dataset=None, defdict=cfg['defdict'],
-                        set_V_SRF=cfg['set_V_SRF'],
-                        lsq_kwargs=cfg['lsq_kwargs'], **cfg['fits_kwargs'])
+                rt1_fits.add_config(
+                    config,
+                    Fits(
+                        dataset=None,
+                        defdict=cfg["defdict"],
+                        set_V_SRF=cfg["set_V_SRF"],
+                        lsq_kwargs=cfg["lsq_kwargs"],
+                        **cfg["fits_kwargs"],
+                    ),
+                )
+
+            # rt1_fits = []
+            # for config in configs:
+            #     cfg = self.get_config(config)
+
+            #     rt1_fits.append(
+            #         [
+            #             config,
+            #             Fits(
+            #                 dataset=None,
+            #                 defdict=cfg["defdict"],
+            #                 set_V_SRF=cfg["set_V_SRF"],
+            #                 lsq_kwargs=cfg["lsq_kwargs"],
+            #                 **cfg["fits_kwargs"],
+            #             ),
+            #         ]
+            #     )
+        else:
+            cfg = self.get_config()
+            rt1_fits = Fits(
+                dataset=None,
+                defdict=cfg["defdict"],
+                set_V_SRF=cfg["set_V_SRF"],
+                lsq_kwargs=cfg["lsq_kwargs"],
+                **cfg["fits_kwargs"],
+            )
 
         return rt1_fits
 
-
     def get_all_modules(self, load_copy=False):
-        '''
+        """
         programmatically import all defined modules from the paths provided
         in the [CONFIGFILES] section
 
@@ -355,22 +475,22 @@ class RT1_configparser(object):
         -------
         processmodules : dict
             a dict with the imported modules
-        '''
+        """
 
         processmodules = dict()
-        for key, val in self.config['CONFIGFILES'].items():
-            if not key.startswith('module__'):
+        for key, val in self.config["CONFIGFILES"].items():
+            if not key.startswith("module__"):
                 continue
 
             modulename = key[8:]
-            processmodules[modulename] = self.get_module(modulename,
-                                                         load_copy=load_copy)
+            processmodules[modulename] = self.get_module(
+                modulename, load_copy=load_copy
+            )
 
         return processmodules
 
-
     def get_module(self, modulename, load_copy=False):
-        '''
+        """
         programmatically import the module 'modulename' as described here:
         https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
 
@@ -388,37 +508,43 @@ class RT1_configparser(object):
         -------
         module : the imported module
 
-        '''
+        """
 
-        assert f'module__{modulename}' in self.config['CONFIGFILES'], (
-            f'the module "{modulename}" is not defined in the [CONFIGFILES]'+
-            ' section of the .ini file     ...available modules are:  '+
-            '  ,  '.join(['"' + i.replace('module__','') + '"'
-                        for i in self.config['CONFIGFILES']
-                        if i.startswith('module__')]))
+        assert f"module__{modulename}" in self.config["CONFIGFILES"], (
+            f'the module "{modulename}" is not defined in the [CONFIGFILES]'
+            + " section of the .ini file     ...available modules are:  "
+            + "  ,  ".join(
+                [
+                    '"' + i.replace("module__", "") + '"'
+                    for i in self.config["CONFIGFILES"]
+                    if i.startswith("module__")
+                ]
+            )
+        )
 
         if load_copy is True:
-            save_path = Path(
-                self.config['PROCESS_SPECS']['path__save_path'].strip())
+            save_path = Path(self.config["PROCESS_SPECS"]["path__save_path"].strip())
 
-            dumpfolder = self.config['PROCESS_SPECS']['dumpfolder'].strip()
+            dumpfolder = self.config["PROCESS_SPECS"]["dumpfolder"].strip()
 
             modulename = Path(
-                self.config['CONFIGFILES'][f'module__{modulename}'].strip()
-                ).name
+                self.config["CONFIGFILES"][f"module__{modulename}"].strip()
+            ).name
 
-            modulepath = save_path / dumpfolder / 'cfg' / modulename
+            modulepath = save_path / dumpfolder / "cfg" / modulename
             assert modulepath.exists(), (
-                f'the module "{modulename}" has not yet been copied to ' +
-                f'"{save_path / dumpfolder / "cfg"}" and can not be ' +
-                'imported using "load_copy=True"')
+                f'the module "{modulename}" has not yet been copied to '
+                + f'"{save_path / dumpfolder / "cfg"}" and can not be '
+                + 'imported using "load_copy=True"'
+            )
         else:
             modulepath = Path(
-                self.config['CONFIGFILES'][f'module__{modulename}'].strip())
+                self.config["CONFIGFILES"][f"module__{modulename}"].strip()
+            )
 
-
-        spec = importlib.util.spec_from_file_location(name=modulepath.stem,
-                                                      location=modulepath)
+        spec = importlib.util.spec_from_file_location(
+            name=modulepath.stem, location=modulepath
+        )
 
         foo = importlib.util.module_from_spec(spec)
         sys.modules[spec.name] = foo
@@ -426,66 +552,85 @@ class RT1_configparser(object):
 
         return foo
 
-
     def get_process_specs(self):
-        inp = self.config['PROCESS_SPECS']
+        """
+        get the process_specs dict
+
+        Returns
+        -------
+        dict
+            the process specs dict.
+
+        """
+        inp = self.config["PROCESS_SPECS"]
 
         process_specs = dict()
         for key, val in inp.items():
-            if key.startswith('datetime__'):
-                date = dict(zip(['s', 'fmt'],
-                        [i.strip() for i in val.split('fmt=')]))
+            # parse None as a 'real' None and not as the string 'None'
+            if val == "None":
+                val = None
+
+            if key.startswith("datetime__"):
+                date = dict(zip(["s", "fmt"], [i.strip() for i in val.split("fmt=")]))
                 process_specs[key[10:]] = self._to_dt(**date)
-            elif key.startswith('path__'):
+            elif key.startswith("path__"):
                 # resolve the path in case .. syntax is used to traverse dirs
-                if '..' in val.strip():
+                if ".." in val.strip():
                     try:
                         process_specs[key[6:]] = Path(val.strip()).resolve()
                     except PermissionError:
-                        print('some paths could not be resolved!',
-                              'avoid using ".." syntax in path-specifications',
-                              ' to circuumvent this issue')
+                        log.warning(
+                            "some paths could not be resolved!"
+                            + 'avoid using ".." syntax in path-specifications'
+                            + " to circuumvent this issue"
+                        )
                         process_specs[key[6:]] = Path(val.strip())
                 else:
                     process_specs[key[6:]] = Path(val.strip())
 
-            elif key.startswith('bool__'):
+            elif key.startswith("bool__"):
                 process_specs[key[6:]] = inp.getboolean(key)
-            elif key.startswith('float__'):
+            elif key.startswith("float__"):
                 process_specs[key[7:]] = inp.getfloat(key)
-            elif key.startswith('int__'):
+            elif key.startswith("int__"):
                 process_specs[key[5:]] = inp.getint(key)
-            elif key.startswith('list__'):
+            elif key.startswith("list__"):
                 listkey = key[6:]
-                if val.startswith('[') and (val.endswith(']')
-                                            or 'fmt=' in val):
+                if val.startswith("[") and (val.endswith("]") or "fmt=" in val):
                     # allow direct conversion of values
-                    if listkey.startswith('bool__'):
+                    if listkey.startswith("bool__"):
                         listkey = listkey[6:]
                         conffunc = ConfigParser()._convert_to_boolean
-                    elif listkey.startswith('float__'):
+                    elif listkey.startswith("float__"):
                         listkey = listkey[7:]
                         conffunc = float
-                    elif listkey.startswith('int__'):
+                    elif listkey.startswith("int__"):
                         listkey = listkey[5:]
                         conffunc = int
-                    elif listkey.startswith('datetime__'):
+                    elif listkey.startswith("datetime__"):
                         listkey = listkey[10:]
-                        spl = dict(zip(['val', 'fmt'],
-                                       [i.strip() for i in val.split('fmt=')]))
-                        if 'fmt' in spl:
-                            conffunc = partial(self._to_dt, fmt=spl['fmt'])
+                        spl = dict(
+                            zip(
+                                ["val", "fmt"],
+                                [i.strip() for i in val.split("fmt=")],
+                            )
+                        )
+                        if "fmt" in spl:
+                            conffunc = partial(self._to_dt, fmt=spl["fmt"])
                         else:
                             conffunc = self._to_dt
-                        val = spl['val']
+                        val = spl["val"]
                     else:
-                        conffunc = lambda x: x
+
+                        def conffunc(x):
+                            return x
+
                     process_specs[listkey] = [
-                        conffunc(i.strip()) for i in
-                        val.strip()[1:-1].split(',')]
+                        conffunc(i.strip()) for i in val.strip()[1:-1].split(",")
+                    ]
 
                 else:
-                    assert False, f'{key} is not a list! (use brackets [] !)'
+                    assert False, f"{key} is not a list! (use brackets [] !)"
             else:
                 process_specs[key] = val
 
