@@ -1,29 +1,43 @@
-"""
-Definition of volume phase scattering functions
-"""
+"""Definition of volume phase scattering functions"""
 
 import numpy as np
 import sympy as sp
-from functools import partial, update_wrapper
+from functools import partial, update_wrapper, lru_cache
 from .scatter import Scatter
 from .rtplots import polarplot
 
 
 class Volume(Scatter):
+    """basic volume class"""
+
     def __init__(self, **kwargs):
-        self.omega = kwargs.pop('omega', None)
-        self.tau = kwargs.pop('tau', None)
+        self.omega = kwargs.pop("omega", None)
+        self.tau = kwargs.pop("tau", None)
 
         # set scattering angle generalization-matrix to [-1,1,1] if it is not
         # explicitly provided by the chosen class this results in a peak in
         # forward-direction which is suitable for describing volume-scattering
         # phase-functions
-        self.a = getattr(self, 'a', [-1., 1., 1.])
+        self.a = getattr(self, "a", [-1.0, 1.0, 1.0])
 
-        #add a quick way for visualizing the functions as polarplot
+        # add a quick way for visualizing the functions as polarplot
         self.polarplot = partial(polarplot, X=self)
         update_wrapper(self.polarplot, polarplot)
 
+    @lru_cache()
+    def _lambda_func(self, *args):
+        # define sympy objects
+        theta_0 = sp.Symbol("theta_0")
+        theta_ex = sp.Symbol("theta_ex")
+        phi_0 = sp.Symbol("phi_0")
+        phi_ex = sp.Symbol("phi_ex")
+
+        # replace arguments and evaluate expression
+        # sp.lambdify is used to allow array-inputs
+        args = (theta_0, theta_ex, phi_0, phi_ex) + tuple(args)
+        pfunc = sp.lambdify(args, self._func, modules=["numpy", "sympy"])
+        return pfunc
+
 
     def p(self, t_0, t_ex, p_0, p_ex, param_dict={}):
         """
@@ -49,35 +63,43 @@ class Volume(Scatter):
         array_like(float)
             Numerical value of the volume-scattering phase-function
         """
-        # define sympy objects
-        theta_0 = sp.Symbol('theta_0')
-        theta_ex = sp.Symbol('theta_ex')
-        phi_0 = sp.Symbol('phi_0')
-        phi_ex = sp.Symbol('phi_ex')
-
-        # replace arguments and evaluate expression
-        # sp.lambdify is used to allow array-inputs
-        # for python > 3.5 unpacking could be used, i.e.:
-        # pfunc = sp.lambdify((theta_0, theta_ex, phi_0, phi_ex,
-        #                     *param_dict.keys()),
-        #                    self._func, modules=["numpy", "sympy"])
-        args = (theta_0, theta_ex, phi_0, phi_ex) + tuple(param_dict.keys())
-        pfunc = sp.lambdify(args, self._func, modules=["numpy", "sympy"])
 
+        # if an explicit numeric function is provided, use it, otherwise
+        # lambdify the available sympy-function
+        if hasattr(self, "_func_numeric"):
+            pfunc = self._func_numeric
+        else:
+            pfunc = self._lambda_func(*param_dict.keys())
         # in case _func is a constant, lambdify will produce a function with
         # scalar output which is not suitable for further processing
         # (this happens e.g. for the Isotropic brdf).
         # The following query is implemented to ensure correct array-output:
         # TODO this is not a proper test !
-        if not isinstance(pfunc(np.array([.1, .2, .3]), .1, .1, .1,
-                                       **{key:.12 for key in param_dict.keys()}
-                                       ), np.ndarray):
+        if not isinstance(
+            pfunc(
+                np.array([0.1, 0.2, 0.3]),
+                0.1,
+                0.1,
+                0.1,
+                **{key: 0.12 for key in param_dict.keys()}
+            ),
+            np.ndarray,
+        ):
             pfunc = np.vectorize(pfunc)
 
         return pfunc(t_0, t_ex, p_0, p_ex, **param_dict)
 
-    def p_theta_diff(self, t_0, t_ex, p_0, p_ex, geometry,
-                param_dict={}, return_symbolic=False, n=1):
+    def p_theta_diff(
+        self,
+        t_0,
+        t_ex,
+        p_0,
+        p_ex,
+        geometry,
+        param_dict={},
+        return_symbolic=False,
+        n=1,
+    ):
         """
         Calculation of the derivative of p with respect to
         the scattering-angles t_ex
@@ -125,7 +147,7 @@ class Volume(Scatter):
             order of derivatives (d^n / d_theta^n)
 
         Returns
-        --------
+        -------
         sympy - expression
             The derivative of the BRDF with espect to the excident angle
             t_ex for the chosen geometry
@@ -133,11 +155,13 @@ class Volume(Scatter):
         """
 
         # define sympy variables based on chosen geometry
-        if geometry == 'mono':
-            assert len(np.unique(p_0)) == 1, 'p_0 must contain only a ' + \
-                'single unique value for monostatic geometry'
+        if geometry == "mono":
+            assert len(np.unique(p_0)) == 1, (
+                "p_0 must contain only a "
+                + "single unique value for monostatic geometry"
+            )
 
-            theta_0 = sp.Symbol('theta_0')
+            theta_0 = sp.Symbol("theta_0")
             theta_ex = theta_0
             phi_0 = np.unique(p_0)[0]
             phi_ex = np.unique(p_0)[0] + sp.pi
@@ -145,83 +169,100 @@ class Volume(Scatter):
             t_ex = t_0
             p_ex = p_0 + np.pi
         else:
-            if geometry[0] == 'v':
-                theta_0 = sp.Symbol('theta_0')
-            elif geometry[0] == 'f':
-                assert len(np.unique(t_0)) == 1, 't_0 must contain only a ' + \
-                    'single unique value for geometry[0] == f'
+            if geometry[0] == "v":
+                theta_0 = sp.Symbol("theta_0")
+            elif geometry[0] == "f":
+                assert len(np.unique(t_0)) == 1, (
+                    "t_0 must contain only a "
+                    + "single unique value for geometry[0] == f"
+                )
 
                 theta_0 = np.unique(t_0)[0]
             else:
-                raise AssertionError('wrong choice of theta_0 geometry')
+                raise AssertionError("wrong choice of theta_0 geometry")
 
-            if geometry[1] == 'v':
-                theta_ex = sp.Symbol('theta_ex')
-            elif geometry[1] == 'f':
-                assert len(np.unique(t_ex)) == 1, 't_ex must contain only' + \
-                    ' a single unique value for geometry[1] == f'
+            if geometry[1] == "v":
+                theta_ex = sp.Symbol("theta_ex")
+            elif geometry[1] == "f":
+                assert len(np.unique(t_ex)) == 1, (
+                    "t_ex must contain only"
+                    + " a single unique value for geometry[1] == f"
+                )
 
                 theta_ex = np.unique(t_ex)[0]
             else:
-                raise AssertionError('wrong choice of theta_ex geometry')
+                raise AssertionError("wrong choice of theta_ex geometry")
 
-            if geometry[2] == 'v':
-                phi_0 = sp.Symbol('phi_0')
-            elif geometry[2] == 'f':
-                assert len(np.unique(p_0)) == 1, 'p_0 must contain only' + \
-                    ' a single unique value for geometry[2] == f'
+            if geometry[2] == "v":
+                phi_0 = sp.Symbol("phi_0")
+            elif geometry[2] == "f":
+                assert len(np.unique(p_0)) == 1, (
+                    "p_0 must contain only"
+                    + " a single unique value for geometry[2] == f"
+                )
 
                 phi_0 = np.unique(p_0)[0]
             else:
-                raise AssertionError('wrong choice of phi_0 geometry')
+                raise AssertionError("wrong choice of phi_0 geometry")
 
-            if geometry[3] == 'v':
-                phi_ex = sp.Symbol('phi_ex')
-            elif geometry[3] == 'f':
-                assert len(np.unique(p_0)) == 1, 'p_ex must contain only' + \
-                    ' a single unique value for geometry[3] == f'
+            if geometry[3] == "v":
+                phi_ex = sp.Symbol("phi_ex")
+            elif geometry[3] == "f":
+                assert len(np.unique(p_0)) == 1, (
+                    "p_ex must contain only"
+                    + " a single unique value for geometry[3] == f"
+                )
 
                 phi_ex = np.unique(p_ex)[0]
             else:
-                raise AssertionError('wrong choice of phi_ex geometry')
+                raise AssertionError("wrong choice of phi_ex geometry")
 
-        if geometry[1] == 'f':
-            dfunc_dtheta_0 = 0.
+        if geometry[1] == "f":
+            dfunc_dtheta_0 = 0.0
         else:
-            func = self._func.xreplace({sp.Symbol('theta_0') : theta_0,
-                                        sp.Symbol('theta_ex') : theta_ex,
-                                        sp.Symbol('phi_0') : phi_0,
-                                        sp.Symbol('phi_ex') : phi_ex,})
+            func = self._func.xreplace(
+                {
+                    sp.Symbol("theta_0"): theta_0,
+                    sp.Symbol("theta_ex"): theta_ex,
+                    sp.Symbol("phi_0"): phi_0,
+                    sp.Symbol("phi_ex"): phi_ex,
+                }
+            )
 
             dfunc_dtheta_0 = sp.diff(func, theta_ex, n)
 
         if return_symbolic is True:
             return dfunc_dtheta_0
         else:
-            args = (sp.Symbol('theta_0'),
-                    sp.Symbol('theta_ex'),
-                    sp.Symbol('phi_0'),
-                    sp.Symbol('phi_ex')) + tuple(param_dict.keys())
+            args = (
+                sp.Symbol("theta_0"),
+                sp.Symbol("theta_ex"),
+                sp.Symbol("phi_0"),
+                sp.Symbol("phi_ex"),
+            ) + tuple(param_dict.keys())
 
+            pfunc = sp.lambdify(args, dfunc_dtheta_0, modules=["numpy", "sympy"])
 
-            pfunc = sp.lambdify(args, dfunc_dtheta_0,
-                                modules=["numpy", "sympy"])
-
-            # in case _func is a constant, lambdify will produce a function with
-            # scalar output which is not suitable for further processing
+            # in case _func is a constant, lambdify will produce a function
+            # with scalar output which is not suitable for further processing
             # (this happens e.g. for the Isotropic brdf).
-            # The following query is implemented to ensure correct array-output:
+            # The following query is implemented to ensure correct array-output
             # TODO this is not a proper test !
-            if not isinstance(pfunc(np.array([.1, .2, .3]), .1, .1, .1,
-                                       **{key:.12 for key in param_dict.keys()}
-                                       ), np.ndarray):
+            if not isinstance(
+                pfunc(
+                    np.array([0.1, 0.2, 0.3]),
+                    0.1,
+                    0.1,
+                    0.1,
+                    **{key: 0.12 for key in param_dict.keys()}
+                ),
+                np.ndarray,
+            ):
                 pfunc = np.vectorize(pfunc)
 
             return pfunc(t_0, t_ex, p_0, p_ex, **param_dict)
 
-
     def legexpansion(self, t_0, t_ex, p_0, p_ex, geometry):
-        assert self.ncoefs > 0
         """
         Definition of the legendre-expansion of the
         volume-scattering phase-function
@@ -275,81 +316,94 @@ class Volume(Scatter):
             (http://rt1.readthedocs.io/en/latest/model_specification.html#evaluation-geometries)
 
         Returns
-        --------
+        -------
         sympy-expression
             The legendre-expansion of the volume-scattering phase-function
             for the chosen geometry
 
         """
-        theta_s = sp.Symbol('theta_s')
-        phi_s = sp.Symbol('phi_s')
+        assert self.ncoefs > 0
+
+        theta_s = sp.Symbol("theta_s")
+        phi_s = sp.Symbol("phi_s")
 
         NP = self.ncoefs
-        n = sp.Symbol('n')
+        n = sp.Symbol("n")
 
         # define sympy variables based on chosen geometry
-        if geometry == 'mono':
+        if geometry == "mono":
 
-            assert len(np.unique(p_0)) == 1, 'p_0 must contain only a ' + \
-                'single unique value for monostatic geometry'
+            assert len(np.unique(p_0)) == 1, (
+                "p_0 must contain only a "
+                + "single unique value for monostatic geometry"
+            )
 
-            theta_0 = sp.Symbol('theta_0')
+            theta_0 = sp.Symbol("theta_0")
             theta_ex = theta_0
             phi_0 = np.unique(p_0)[0]
             phi_ex = np.unique(p_0)[0] + sp.pi
         else:
-            if geometry[0] == 'v':
-                theta_0 = sp.Symbol('theta_0')
-            elif geometry[0] == 'f':
-                assert len(np.unique(t_0)) == 1, 't_0 must contain only a ' + \
-                    'single unique value for geometry[0] == f'
+            if geometry[0] == "v":
+                theta_0 = sp.Symbol("theta_0")
+            elif geometry[0] == "f":
+                assert len(np.unique(t_0)) == 1, (
+                    "t_0 must contain only a "
+                    + "single unique value for geometry[0] == f"
+                )
 
                 theta_0 = np.unique(t_0)[0]
             else:
-                raise AssertionError('wrong choice of theta_i geometry')
+                raise AssertionError("wrong choice of theta_i geometry")
 
-            if geometry[1] == 'v':
-                theta_ex = sp.Symbol('theta_ex')
-            elif geometry[1] == 'f':
-                assert len(np.unique(t_ex)) == 1, 't_ex must contain only' + \
-                    ' a single unique value for geometry[1] == f'
+            if geometry[1] == "v":
+                theta_ex = sp.Symbol("theta_ex")
+            elif geometry[1] == "f":
+                assert len(np.unique(t_ex)) == 1, (
+                    "t_ex must contain only"
+                    + " a single unique value for geometry[1] == f"
+                )
 
                 theta_ex = np.unique(t_ex)[0]
             else:
-                raise AssertionError('wrong choice of theta_ex geometry')
+                raise AssertionError("wrong choice of theta_ex geometry")
 
-            if geometry[2] == 'v':
-                phi_0 = sp.Symbol('phi_0')
-            elif geometry[2] == 'f':
-                assert len(np.unique(p_0)) == 1, 'p_0 must contain only' + \
-                    ' a single unique value for geometry[2] == f'
+            if geometry[2] == "v":
+                phi_0 = sp.Symbol("phi_0")
+            elif geometry[2] == "f":
+                assert len(np.unique(p_0)) == 1, (
+                    "p_0 must contain only"
+                    + " a single unique value for geometry[2] == f"
+                )
 
                 phi_0 = np.unique(p_0)[0]
             else:
-                raise AssertionError('wrong choice of phi_i geometry')
+                raise AssertionError("wrong choice of phi_i geometry")
 
-            if geometry[3] == 'v':
-                phi_ex = sp.Symbol('phi_ex')
-            elif geometry[3] == 'f':
-                assert len(np.unique(p_0)) == 1, 'p_ex must contain only' + \
-                    ' a single unique value for geometry[3] == f'
+            if geometry[3] == "v":
+                phi_ex = sp.Symbol("phi_ex")
+            elif geometry[3] == "f":
+                assert len(np.unique(p_0)) == 1, (
+                    "p_ex must contain only"
+                    + " a single unique value for geometry[3] == f"
+                )
 
                 phi_ex = np.unique(p_ex)[0]
             else:
-                raise AssertionError('wrong choice of phi_ex geometry')
+                raise AssertionError("wrong choice of phi_ex geometry")
 
         # correct for backscattering
-        return sp.Sum(self.legcoefs *
-                      sp.legendre(n,
-                                  self.scat_angle(sp.pi - theta_0,
-                                                  theta_s,
-                                                  phi_0,
-                                                  phi_s,
-                                                  self.a)), (n, 0, NP - 1))
+        return sp.Sum(
+            self.legcoefs
+            * sp.legendre(
+                n,
+                self.scat_angle(sp.pi - theta_0, theta_s, phi_0, phi_s, self.a),
+            ),
+            (n, 0, NP - 1),
+        )
 
 
 class LinCombV(Volume):
-    '''
+    """
     Class to generate linear-combinations of volume-class elements
 
     For details please look at the documentation
@@ -377,32 +431,28 @@ class LinCombV(Volume):
                a list that contains the the individual phase-functions
                (Volume-objects) and the associated weighting-factors
                (floats) of the linear-combination.
-    '''
+    """
 
     def __init__(self, Vchoices=None, **kwargs):
         super(LinCombV, self).__init__(**kwargs)
 
         self.Vchoices = Vchoices
-        self._set_function()
-        self._set_legexpansion()
 
-    def _set_function(self):
-        """
-        define phase function as sympy object for later evaluation
-        """
+    @property
+    @lru_cache()
+    def _func(self):
+        """define phase function as sympy object for later evaluation"""
 
-        self._func = self._Vcombiner()._func
+        return self._Vcombiner()._func
 
     def _set_legexpansion(self):
-        '''
-        set legexpansion to the combined legexpansion
-        '''
+        """set legexpansion to the combined legexpansion"""
 
         self.ncoefs = self._Vcombiner().ncoefs
         self.legexpansion = self._Vcombiner().legexpansion
 
     def _Vcombiner(self):
-        '''
+        """
         Returns a Volume-class element based on an input-array of Volume-class
         elements. The array must be shaped in the form:
             Vchoices = [  [ weighting-factor   ,   Volume-class element ]  ,
@@ -412,7 +462,8 @@ class LinCombV(Volume):
         In order to keep the normalization of the phase-functions correct,
         the sum of the weighting factors must equate to 1!
 
-        ATTENTION:
+        Attention
+        ---------
             the .legexpansion()-function of the combined volume-class
             element is no longer related to its legcoefs (which are set to 0.)
             since the individual legexpansions of the combined volume-class
@@ -420,7 +471,7 @@ class LinCombV(Volume):
             of the generalized scattering angle! This does not affect any
             calculations, since the evaluation is exclusively based on the
             use of the .legexpansion()-function.
-        '''
+        """
 
         class Phasefunction(Volume):
             """
@@ -430,29 +481,18 @@ class LinCombV(Volume):
 
             def __init__(self, **kwargs):
                 super(Phasefunction, self).__init__(**kwargs)
-                self._set_function()
-                self._set_legcoefficients()
-
-            def _set_function(self):
-                """
-                define phase function as sympy object for later evaluation
-                """
-
-                self._func = 0.
-
-            def _set_legcoefficients(self):
-                """
-                set Legrende coefficients
-                needs to be a function that can be later evaluated by
-                subsituting 'n'
-                """
-
-                self.legcoefs = 0.
+                self._func = 0.0
+                self.legcoefs = 0.0
 
         # find phase functions with equal a parameters
-        equals = [np.where((np.array([VV[1].a for VV in self.Vchoices]) ==
-                            tuple(V[1].a)).all(axis=1))[0]
-                  for V in self.Vchoices]
+        equals = [
+            np.where(
+                (np.array([VV[1].a for VV in self.Vchoices]) == tuple(V[1].a)).all(
+                    axis=1
+                )
+            )[0]
+            for V in self.Vchoices
+        ]
 
         # evaluate index of phase-functions that have equal a parameter
         equal_a = list({tuple(row) for row in equals})
@@ -490,8 +530,8 @@ class LinCombV(Volume):
         # combine legendre-expansions for each a-parameter based on given
         # combined legendre-coefficients
         Vcomb.legexpansion = lambda t_0, t_ex, p_0, p_ex, geometry: np.sum(
-            [lexp(t_0, t_ex, p_0, p_ex, geometry)
-             for lexp in dummylegexpansion])
+            [lexp(t_0, t_ex, p_0, p_ex, geometry) for lexp in dummylegexpansion]
+        )
 
         for V in self.Vchoices:
             # set parameters based on chosen classes to define analytic
@@ -506,7 +546,7 @@ class Rayleigh(Volume):
     Define a Rayleigh scattering function
 
     Parameters
-    -----------
+    ----------
     tau : scalar(float)
           Optical depth
 
@@ -524,34 +564,42 @@ class Rayleigh(Volume):
 
     def __init__(self, **kwargs):
         super(Rayleigh, self).__init__(**kwargs)
-        self._set_function()
-        self._set_legcoefficients()
 
-    def _set_function(self):
-        """
-        define phase function as sympy object for later evaluation
-        """
-        theta_0 = sp.Symbol('theta_0')
-        theta_ex = sp.Symbol('theta_ex')
-        phi_0 = sp.Symbol('phi_0')
-        phi_ex = sp.Symbol('phi_ex')
+    @property
+    def ncoefs(self):
+        # make ncoefs a property since it is fixed and should not be changed
+        # only 3 coefficients are needed to correctly represent
+        # the Rayleigh scattering function
+        return 3
+
+    @property
+    @lru_cache()
+    def _func(self):
+        """define phase function as sympy object for later evaluation"""
+        theta_0 = sp.Symbol("theta_0")
+        theta_ex = sp.Symbol("theta_ex")
+        phi_0 = sp.Symbol("phi_0")
+        phi_ex = sp.Symbol("phi_ex")
         x = self.scat_angle(theta_0, theta_ex, phi_0, phi_ex, self.a)
-        self._func = 3. / (16. * sp.pi) * (1. + x ** 2.)
+        return 3.0 / (16.0 * sp.pi) * (1.0 + x ** 2.0)
 
-    def _set_legcoefficients(self):
+    @property
+    @lru_cache()
+    def legcoefs(self):
         """
         set Legrende coefficients
         needs to be a function that can be later evaluated by subsituting 'n'
         """
         # only 3 coefficients are needed to correctly represent
         # the Rayleigh scattering function
-        self.ncoefs = 3
-        n = sp.Symbol('n')
-        self.legcoefs = ((3. / (16. * sp.pi)) * ((4. / 3.) *
-                                                 sp.KroneckerDelta(0, n) +
-                                                 (2. / 3.) *
-                                                 sp.KroneckerDelta(2, n))
-                         ).expand()
+        n = sp.Symbol("n")
+        return (
+            (3.0 / (16.0 * sp.pi))
+            * (
+                (4.0 / 3.0) * sp.KroneckerDelta(0, n)
+                + (2.0 / 3.0) * sp.KroneckerDelta(2, n)
+            )
+        ).expand()
 
 
 class HenyeyGreenstein(Volume):
@@ -559,7 +607,7 @@ class HenyeyGreenstein(Volume):
     Define a HenyeyGreenstein scattering function
 
     Parameters
-    -----------
+    ----------
     tau : scalar(float)
           Optical depth
 
@@ -578,41 +626,59 @@ class HenyeyGreenstein(Volume):
         (http://rt1.readthedocs.io/en/latest/theory.html#equation-general_scat_angle)
     """
 
-    def __init__(self, t=None, ncoefs=None, a=[-1., 1., 1.], **kwargs):
-        assert t is not None, 't parameter needs to be provided!'
-        assert ncoefs is not None, 'Number of coeffs needs to be specified'
+    def __init__(self, t=None, ncoefs=None, a=[-1.0, 1.0, 1.0], **kwargs):
+        assert t is not None, "t parameter needs to be provided!"
+        assert ncoefs is not None, "Number of coeffs needs to be specified"
         super(HenyeyGreenstein, self).__init__(**kwargs)
         self.t = t
         self.a = a
-        assert isinstance(self.a, list), 'Error: Generalization-parameter ' + \
-            'needs to be a list'
-        assert len(a) == 3, 'Error: Generalization-parameter list must ' + \
-            'contain 3 values'
+        assert isinstance(self.a, list), (
+            "Error: Generalization-parameter " + "needs to be a list"
+        )
+        assert len(a) == 3, (
+            "Error: Generalization-parameter list must " + "contain 3 values"
+        )
         self.ncoefs = ncoefs
         assert self.ncoefs > 0
-        self._set_function()
-        self._set_legcoefficients()
 
-    def _set_function(self):
-        """
-        define phase function as sympy object for later evaluation
-        """
-        theta_0 = sp.Symbol('theta_0')
-        theta_ex = sp.Symbol('theta_ex')
-        phi_0 = sp.Symbol('phi_0')
-        phi_ex = sp.Symbol('phi_ex')
+    @property
+    @lru_cache()
+    def _func(self):
+        """define phase function as sympy object for later evaluation"""
+        theta_0 = sp.Symbol("theta_0")
+        theta_ex = sp.Symbol("theta_ex")
+        phi_0 = sp.Symbol("phi_0")
+        phi_ex = sp.Symbol("phi_ex")
         x = self.scat_angle(theta_0, theta_ex, phi_0, phi_ex, self.a)
-        self._func = (1. - self.t ** 2.) / ((4. * sp.pi) *
-                                            (1. + self.t ** 2. -
-                                             2. * self.t * x) ** 1.5)
+        func = (1.0 - self.t ** 2.0) / (
+            (4.0 * sp.pi) * (1.0 + self.t ** 2.0 - 2.0 * self.t * x) ** 1.5
+        )
 
-    def _set_legcoefficients(self):
+        return func
+
+    def _func_numeric(self, theta_0, theta_ex, phi_0, phi_ex, **kwargs):
+        """direct numeric version of _func"""
+        if isinstance(self.t, sp.Symbol):
+            t = kwargs[str(self.t)]
+        else:
+            t = self.t
+        x = self._scat_angle_numeric(theta_0, theta_ex, phi_0, phi_ex, self.a)
+        func = (1.0 - t ** 2.0) / (
+            (4.0 * np.pi) * (1.0 + t ** 2.0 - 2.0 * t * x) ** 1.5
+        )
+
+        return func
+
+    @property
+    @lru_cache()
+    def legcoefs(self):
         """
         set Legrende coefficients
         needs to be a function that can be later evaluated by subsituting 'n'
         """
-        n = sp.Symbol('n')
-        self.legcoefs = (1. / (4. * sp.pi)) * (2. * n + 1) * self.t ** n
+        n = sp.Symbol("n")
+        legcoefs = (1.0 / (4.0 * sp.pi)) * (2.0 * n + 1) * self.t ** n
+        return legcoefs
 
 
 class HGRayleigh(Volume):
@@ -624,7 +690,7 @@ class HGRayleigh(Volume):
         Appl. Opt., 45(28):7475-7479, Oct 2006. doi: 10.1364/AO.45.'
 
     Parameters
-    -----------
+    ----------
     tau : scalar(float)
           Optical depth
 
@@ -643,55 +709,78 @@ class HGRayleigh(Volume):
         (http://rt1.readthedocs.io/en/latest/theory.html#equation-general_scat_angle)
     """
 
-    def __init__(self, t=None, ncoefs=None, a=[-1., 1., 1.], **kwargs):
-        assert t is not None, 't parameter needs to be provided!'
-        assert ncoefs is not None, 'Number of coeffs needs to be specified'
+    def __init__(self, t=None, ncoefs=None, a=[-1.0, 1.0, 1.0], **kwargs):
+        assert t is not None, "t parameter needs to be provided!"
+        assert ncoefs is not None, "Number of coeffs needs to be specified"
         super(HGRayleigh, self).__init__(**kwargs)
         self.t = t
         self.a = a
-        assert isinstance(self.a, list), 'Error: Generalization-parameter ' + \
-            'needs to be a list'
-        assert len(a) == 3, 'Error: Generalization-parameter list must ' + \
-            'contain 3 values'
-        assert all(type(x) == float for x in a), 'Error: Generalization-' + \
-            'parameter array must contain only floating-point values!'
+        assert isinstance(self.a, list), (
+            "Error: Generalization-parameter " + "needs to be a list"
+        )
+        assert len(a) == 3, (
+            "Error: Generalization-parameter list must " + "contain 3 values"
+        )
+        assert all(type(x) == float for x in a), (
+            "Error: Generalization-"
+            + "parameter array must contain only floating-point values!"
+        )
         self.ncoefs = ncoefs
         assert self.ncoefs > 0
-        self._set_function()
-        self._set_legcoefficients()
 
-    def _set_function(self):
-        """
-        define phase function as sympy object for later evaluation
-        """
-        theta_0 = sp.Symbol('theta_0')
-        theta_ex = sp.Symbol('theta_ex')
-        phi_0 = sp.Symbol('phi_0')
-        phi_ex = sp.Symbol('phi_ex')
+    @property
+    @lru_cache()
+    def _func(self):
+        """define phase function as sympy object for later evaluation"""
+        theta_0 = sp.Symbol("theta_0")
+        theta_ex = sp.Symbol("theta_ex")
+        phi_0 = sp.Symbol("phi_0")
+        phi_ex = sp.Symbol("phi_ex")
         x = self.scat_angle(theta_0, theta_ex, phi_0, phi_ex, self.a)
-        self._func = 3. / (8. * sp.pi) * (1. /
-                                          (2. + self.t ** 2) *
-                                          (1 + x ** 2) *
-                                          (1. - self.t ** 2.) /
-                                          ((1. + self.t ** 2. -
-                                            2. * self.t * x) ** 1.5))
+        return (
+            3.0
+            / (8.0 * sp.pi)
+            * (
+                1.0
+                / (2.0 + self.t ** 2)
+                * (1 + x ** 2)
+                * (1.0 - self.t ** 2.0)
+                / ((1.0 + self.t ** 2.0 - 2.0 * self.t * x) ** 1.5)
+            )
+        )
 
-    def _set_legcoefficients(self):
+    @property
+    @lru_cache()
+    def legcoefs(self):
         """
         set Legrende coefficients
         needs to be a function that can be later evaluated by subsituting 'n'
         """
-        n = sp.Symbol('n')
-        self.legcoefs = sp.Piecewise(
-            (3. / (8. * sp.pi) * 1. / (2. + self.t ** 2) *
-             ((n + 2.) * (n + 1.) /
-              (2. * n + 3) * self.t ** (n + 2.) +
-              (n + 1.) ** 2. / (2. * n + 3.) * self.t ** n +
-              (5. * n ** 2. - 1.) / (2. * n - 1.) * self.t ** n), n < 2),
-            (3. / (8. * sp.pi) * 1. / (2. + self.t ** 2) *
-             (n * (n - 1.) / (2. * n - 1.) * self.t ** (n - 2.) +
-              (n + 2.) * (n + 1.) / (2. * n + 3) * self.t ** (n + 2.) +
-              (n + 1.) ** 2. / (2. * n + 3.) * self.t ** n +
-              (5. * n ** 2. - 1.) / (2. * n - 1.) * self.t ** n), True)
+        n = sp.Symbol("n")
+        return sp.Piecewise(
+            (
+                3.0
+                / (8.0 * sp.pi)
+                * 1.0
+                / (2.0 + self.t ** 2)
+                * (
+                    (n + 2.0) * (n + 1.0) / (2.0 * n + 3) * self.t ** (n + 2.0)
+                    + (n + 1.0) ** 2.0 / (2.0 * n + 3.0) * self.t ** n
+                    + (5.0 * n ** 2.0 - 1.0) / (2.0 * n - 1.0) * self.t ** n
+                ),
+                n < 2,
+            ),
+            (
+                3.0
+                / (8.0 * sp.pi)
+                * 1.0
+                / (2.0 + self.t ** 2)
+                * (
+                    n * (n - 1.0) / (2.0 * n - 1.0) * self.t ** (n - 2.0)
+                    + (n + 2.0) * (n + 1.0) / (2.0 * n + 3) * self.t ** (n + 2.0)
+                    + (n + 1.0) ** 2.0 / (2.0 * n + 3.0) * self.t ** n
+                    + (5.0 * n ** 2.0 - 1.0) / (2.0 * n - 1.0) * self.t ** n
+                ),
+                True,
+            ),
         )
-
